@@ -6,14 +6,14 @@ clientbot.py: Clientbot (regular IRC bot) protocol module for PyLink.
 # that a regular server would have (e.g. spawning virtual users for things like Relay). Somehow it
 # works on most networks though!
 
-import time
-import threading
 import base64
+import threading
+import time
 
-from pylinkirc import utils, conf
+from pylinkirc import conf, utils
+from pylinkirc.classes import *
 from pylinkirc.log import log
 from pylinkirc.protocols.ircs2s_common import *
-from pylinkirc.classes import *
 
 FALLBACK_REALNAME = 'PyLink Relay Mirror Client'
 
@@ -21,18 +21,198 @@ FALLBACK_REALNAME = 'PyLink Relay Mirror Client'
 IRCV3_CAPABILITIES = {'multi-prefix', 'sasl', 'away-notify', 'userhost-in-names', 'chghost', 'account-notify',
                       'account-tag', 'extended-join'}
 
-class ClientbotWrapperProtocol(IRCCommonProtocol):
+class ClientbotBaseProtocol(PyLinkNetworkCoreWithUtils):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
-        self.protocol_caps = {'visible-state-only', 'slash-in-nicks', 'slash-in-hosts', 'underscore-in-hosts',
-                              'ssl-should-verify'}
-
-        self.has_eob = False
+        self.protocol_caps |= {'visible-state-only', 'slash-in-nicks', 'slash-in-hosts', 'underscore-in-hosts',
+                               'freeform-nicks'}
 
         # Remove conf key checks for those not needed for Clientbot.
         self.conf_keys -= {'recvpass', 'sendpass', 'sid', 'sidrange', 'hostname'}
 
+    def _get_UID(self, nick, ident=None, host=None, spawn_new=False):
+        """
+        Fetches the UID for the given nick, creating one if it does not already exist and spawn_new is True.
+
+        To prevent message spoofing, this will only return an external (non-PyLink) client or the PyLink bot itself.
+        """
+        #log.debug('(%s) _get_UID: searching for nick %s', self.name, nick, stack_info=True)
+        idsource = self.nick_to_uid(nick, filterfunc=lambda uid: uid == self.pseudoclient.uid or not self.is_internal_client(uid))
+
+        if idsource is None and spawn_new:
+            # If this sender doesn't already exist, spawn a new client.
+            idsource = self.spawn_client(nick, ident or 'unknown', host or 'unknown',
+                                         server=self.uplink, realname=FALLBACK_REALNAME).uid
+        return idsource or nick  # Return input if missing per upstream spec
+
+    def away(self, source, text):
+        """STUB: sets away messages for internal clients."""
+        log.debug('(%s) away: target is %s, internal client? %s', self.name, source, self.is_internal_client(source))
+
+        if self.users[source].away != text:
+            if not self.is_internal_client(source):
+                log.debug('(%s) away: sending AWAY hook from %s with text %r', self.name, source, text)
+                self.call_hooks([source, 'AWAY', {'text': text}])
+
+            self.users[source].away = text
+
+    def join(self, client, channel):
+        """STUB: sends a virtual join (CLIENTBOT_JOIN) from the client to channel."""
+        self._channels[channel].users.add(client)
+        self.users[client].channels.add(channel)
+
+        if self.pseudoclient and client != self.pseudoclient:
+            log.debug('(%s) join: faking JOIN of client %s/%s to %s', self.name, client,
+                      self.get_friendly_name(client), channel)
+            self.call_hooks([client, 'CLIENTBOT_JOIN', {'channel': channel}])
+
+    def kick(self, source, channel, target, reason=''):
+        """STUB: rejoins users on kick attempts, for server transports where kicking users from channels is not implemented."""
+        if not self.is_internal_client(target):
+            log.info("(%s) Rejoining user %s to %s since kicks are not supported here.", self.name, target, channel)
+            self.join(target, channel)
+            self.call_hooks([None, 'JOIN', {'channel': channel, 'users': [target], 'modes': []}])
+        elif channel in self.channels:
+            self.channels[channel].remove_user(target)
+            self.users[target].channels.discard(channel)
+            self.call_hooks([source, 'CLIENTBOT_KICK', {'channel': channel, 'target': target, 'text': reason}])
+        else:
+            log.warning('(%s) Possible desync? Tried to kick() on non-existent channel %s', self.name, channel)
+
+    def message(self, source, target, text, notice=False):
+        """STUB: Sends messages to the target."""
+        if self.pseudoclient and self.pseudoclient.uid != source:
+            self.call_hooks([source, 'CLIENTBOT_MESSAGE', {'target': target, 'is_notice': notice, 'text': text}])
+
+    def nick(self, source, newnick):
+        """STUB: sends a virtual nick change (CLIENTBOT_NICK)."""
+        assert source, "No source given?"
+        self.call_hooks([source, 'CLIENTBOT_NICK', {'newnick': newnick}])
+        self.users[source].nick = newnick
+
+    def notice(self, source, target, text):
+        """Sends notices to the target."""
+        # Wrap around message(), which does all the text formatting for us.
+        self.message(source, target, text, notice=True)
+
+    def sjoin(self, server, channel, users, ts=None, modes=set()):
+        """STUB: bursts joins from a server."""
+        # This stub only updates the state internally with the users given. modes and TS are currently ignored.
+        puids = {u[-1] for u in users}
+        for user in puids:
+            self.users[user].channels.add(channel)
+
+        self._channels[channel].users |= puids
+        nicks = {self.get_friendly_name(u) for u in puids}
+        self.call_hooks([server, 'CLIENTBOT_SJOIN', {'channel': channel, 'nicks': nicks}])
+
+    # Note: clientbot clients are initialized with umode +i by default
+    def spawn_client(self, nick, ident='unknown', host='unknown.host', realhost=None, modes={('i', None)},
+            server=None, ip='0.0.0.0', realname='', ts=None, opertype=None,
+            manipulatable=False):
+        """
+        STUB: Pretends to spawn a new client with a subset of the given options.
+        """
+
+        server = server or self.sid
+        uid = self.uidgen.next_uid(prefix=nick)
+
+        ts = ts or int(time.time())
+
+        log.debug('(%s) spawn_client stub called, saving nick %s as PUID %s', self.name, nick, uid)
+        u = self.users[uid] = User(self, nick, ts, uid, server, ident=ident, host=host, realname=realname,
+                                   manipulatable=manipulatable, realhost=realhost, ip=ip)
+        self.servers[server].users.add(uid)
+
+        self.apply_modes(uid, modes)
+
+        return u
+
+    def spawn_server(self, name, sid=None, uplink=None, desc=None, internal=True):
+        """
+        STUB: Pretends to spawn a new server with a subset of the given options.
+        """
+        if internal:
+            # Use a custom pseudo-SID format for internal servers to prevent any server name clashes
+            sid = self.sidgen.next_sid(prefix=name)
+        else:
+            # For others servers, just use the server name as the SID.
+            sid = name
+
+        self.servers[sid] = Server(self, uplink, name, internal=internal)
+        return sid
+
+    def squit(self, source, target, text):
+        """STUB: SQUITs a server."""
+        # What this actually does is just handle the SQUIT internally: i.e.
+        # Removing pseudoclients and pseudoservers.
+        squit_data = self._squit(source, 'CLIENTBOT_VIRTUAL_SQUIT', [target, text])
+
+        if squit_data and squit_data.get('nicks'):
+            self.call_hooks([source, 'CLIENTBOT_SQUIT', squit_data])
+
+    def part(self, source, channel, reason=''):
+        """STUB: Parts a user from a channel."""
+        if self.pseudoclient and source == self.pseudoclient.uid:
+            raise NotImplementedError("Explicitly leaving channels is not supported here.")
+        self._channels[channel].remove_user(source)
+        self.users[source].channels.discard(channel)
+        self.call_hooks([source, 'CLIENTBOT_PART', {'channel': channel, 'text': reason}])
+
+    def quit(self, source, reason):
+        """STUB: Quits a client."""
+        userdata = self._remove_client(source)
+        self.call_hooks([source, 'CLIENTBOT_QUIT', {'text': reason, 'userdata': userdata}])
+
+    def _stub(self, *args):
+        """Stub outgoing command function (does nothing)."""
+        return
+    # Note: invite() and mode() are implemented in ClientbotWrapperProtocol below
+    invite = mode = topic = topic_burst = _stub  # XXX: incomplete
+
+    def _stub_raise(self, *args):
+        """Stub outgoing command function (raises an error)."""
+        raise NotImplementedError("Not supported on Clientbot")
+    kill = knock = numeric = _stub_raise
+
+    def update_client(self, target, field, text):
+        """Updates the known ident, host, or realname of a client."""
+        # Note: unlike other protocol modules, this function is also called as a helper to
+        # update data for external clients.
+        # Following this, we only want to send hook payloads if the target is an external client.
+        if target not in self.users:
+            log.warning("(%s) Unknown target %s for update_client()", self.name, target)
+            return
+
+        u = self.users[target]
+
+        if field == 'IDENT' and u.ident != text:
+            u.ident = text
+            if not self.is_internal_client(target):
+                self.call_hooks([self.sid, 'CHGIDENT',
+                                   {'target': target, 'newident': text}])
+        elif field == 'HOST' and u.host != text:
+            u.host = text
+            if not self.is_internal_client(target):
+                self.call_hooks([self.sid, 'CHGHOST',
+                                   {'target': target, 'newhost': text}])
+        elif field in ('REALNAME', 'GECOS') and u.realname != text:
+            u.realname = text
+            if not self.is_internal_client(target):
+                self.call_hooks([self.sid, 'CHGNAME',
+                                   {'target': target, 'newgecos': text}])
+        else:
+            return  # Nothing changed
+
+class ClientbotWrapperProtocol(ClientbotBaseProtocol, IRCCommonProtocol):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        self.protocol_caps |= {'ssl-should-verify'}
+
+        self.has_eob = False
+
         # This is just a fallback. Actual casemapping is fetched by handle_005()
         self.casemapping = 'ascii'
 
@@ -109,52 +289,6 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
                                  ident=ident, realname=realname, host=self.hostname())
         self.users[self.pseudoclient.uid] = self.pseudoclient
 
-    # Note: clientbot clients are initialized with umode +i by default
-    def spawn_client(self, nick, ident='unknown', host='unknown.host', realhost=None, modes={('i', None)},
-            server=None, ip='0.0.0.0', realname='', ts=None, opertype=None,
-            manipulatable=False):
-        """
-        STUB: Pretends to spawn a new client with a subset of the given options.
-        """
-
-        server = server or self.sid
-        uid = self.uidgen.next_uid(prefix=nick)
-
-        ts = ts or int(time.time())
-
-        log.debug('(%s) spawn_client stub called, saving nick %s as PUID %s', self.name, nick, uid)
-        u = self.users[uid] = User(self, nick, ts, uid, server, ident=ident, host=host, realname=realname,
-                                   manipulatable=manipulatable, realhost=realhost, ip=ip)
-        self.servers[server].users.add(uid)
-
-        self.apply_modes(uid, modes)
-
-        return u
-
-    def spawn_server(self, name, sid=None, uplink=None, desc=None, internal=True):
-        """
-        STUB: Pretends to spawn a new server with a subset of the given options.
-        """
-        if internal:
-            # Use a custom pseudo-SID format for internal servers to prevent any server name clashes
-            sid = self.sidgen.next_sid(prefix=name)
-        else:
-            # For others servers, just use the server name as the SID.
-            sid = name
-
-        self.servers[sid] = Server(self, uplink, name, internal=internal)
-        return sid
-
-    def away(self, source, text):
-        """STUB: sets away messages for clients internally."""
-        log.debug('(%s) away: target is %s, internal client? %s', self.name, source, self.is_internal_client(source))
-
-        if self.users[source].away != text:
-            if not self.is_internal_client(source):
-                log.debug('(%s) away: sending AWAY hook from %s with text %r', self.name, source, text)
-                self.call_hooks([source, 'AWAY', {'text': text}])
-
-            self.users[source].away = text
 
     def invite(self, client, target, channel):
         """Invites a user to a channel."""
@@ -169,12 +303,8 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         if self.pseudoclient and client == self.pseudoclient.uid:
             self.send('JOIN %s' % channel)
         else:
-            self._channels[channel].users.add(client)
-            self.users[client].channels.add(channel)
-
-            log.debug('(%s) join: faking JOIN of client %s/%s to %s', self.name, client,
-                      self.get_friendly_name(client), channel)
-            self.call_hooks([client, 'CLIENTBOT_JOIN', {'channel': channel}])
+            # Pass on a virtual JOIN as a hook
+            super().join(client, channel)
 
     def kick(self, source, channel, target, reason=''):
         """Sends channel kicks."""
@@ -191,7 +321,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
             return
 
         self.send('KICK %s %s :<%s> %s' % (channel, self._expandPUID(target),
-                      self.get_friendly_name(source), reason))
+                  self.get_friendly_name(source), reason))
 
         # Don't update our state here: wait for the IRCd to send an acknowledgement instead.
         # There is essentially a 3 second wait to do this, as we send NAMES with a delay
@@ -215,7 +345,8 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         if self.pseudoclient and self.pseudoclient.uid == source:
             self.send('%s %s :%s' % (command, self._expandPUID(target), text))
         else:
-            self.call_hooks([source, 'CLIENTBOT_MESSAGE', {'target': target, 'is_notice': notice, 'text': text}])
+            # Pass the message on as a hook
+            super().message(source, target, text, notice=notice)
 
     def mode(self, source, channel, modes, ts=None):
         """Sends channel MODE changes."""
@@ -244,7 +375,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
             log.debug('(%s) mode: filtered modes for %s: %s', self.name, channel, extmodes)
             if extmodes:
                 bufsize = self.S2S_BUFSIZE - len(':%s MODE %s ' % (self.get_hostmask(self.pseudoclient.uid), channel))
-                for msg in self.wrap_modes(extmodes, bufsize, max_modes_per_msg=int(self._caps.get('MODES', 0))):
+                for msg in self.wrap_modes(extmodes, bufsize, max_modes_per_msg=int(self._caps.get('MODES') or 0)):
                     self.send('MODE %s %s' % (channel, msg))
                     # Don't update the state here: the IRCd sill respond with a MODE reply if successful.
 
@@ -254,20 +385,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
             self.send('NICK :%s' % newnick)
             # No state update here: the IRCd will respond with a NICK acknowledgement if the change succeeds.
         else:
-            assert source, "No source given?"
-            # Check that the new nick exists and isn't the same client as the sender
-            # (for changing nick case)
-            nick_uid = self.nick_to_uid(newnick)
-            if nick_uid and nick_uid != source:
-                log.warning('(%s) Blocking attempt from virtual client %s to change nick to %s (nick in use)', self.name, source, newnick)
-                return
-            self.call_hooks([source, 'CLIENTBOT_NICK', {'newnick': newnick}])
-            self.users[source].nick = newnick
-
-    def notice(self, source, target, text):
-        """Sends notices to the target."""
-        # Wrap around message(), which does all the text formatting for us.
-        self.message(source, target, text, notice=True)
+            super().nick(source, newnick)
 
     def _ping_uplink(self):
         """
@@ -280,6 +398,13 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
             for channel in self.pseudoclient.channels:
                 self._send_who(channel)
 
+            # Join persistent channels if always_autorejoin is enabled and there are any we're not in
+            if self.serverdata.get('always_autorejoin') and self.has_cap('can-manage-bot-channels'):
+                for channel in world.services['pylink'].get_persistent_channels(self):
+                    if channel not in self.pseudoclient.channels:
+                        log.info('(%s) Attempting to rejoin %s', self.name, channel)
+                        self.join(self.pseudoclient.uid, channel)
+
     def part(self, source, channel, reason=''):
         """STUB: Parts a user from a channel."""
         self._channels[channel].remove_user(source)
@@ -290,13 +415,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
             self._channels[channel]._clientbot_part_requested = True
             self.send('PART %s :%s' % (channel, reason))
         else:
-            self.call_hooks([source, 'CLIENTBOT_PART', {'channel': channel, 'text': reason}])
-
-    def quit(self, source, reason):
-        """STUB: Quits a client."""
-        userdata = self.users[source]
-        self._remove_client(source)
-        self.call_hooks([source, 'CLIENTBOT_QUIT', {'text': reason, 'userdata': userdata}])
+            super().part(source, channel, reason=reason)
 
     def sjoin(self, server, channel, users, ts=None, modes=set()):
         """STUB: bursts joins from a server."""
@@ -315,99 +434,6 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         nicks = {self.get_friendly_name(u) for u in puids}
         self.call_hooks([server, 'CLIENTBOT_SJOIN', {'channel': channel, 'nicks': nicks}])
 
-    def squit(self, source, target, text):
-        """STUB: SQUITs a server."""
-        # What this actually does is just handle the SQUIT internally: i.e.
-        # Removing pseudoclients and pseudoservers.
-        squit_data = self._squit(source, 'CLIENTBOT_VIRTUAL_SQUIT', [target, text])
-
-        if squit_data and squit_data.get('nicks'):
-            self.call_hooks([source, 'CLIENTBOT_SQUIT', squit_data])
-
-    def _stub(self, *args):
-        """Stub outgoing command function (does nothing)."""
-        return
-    topic = topic_burst = _stub  # XXX: incomplete
-
-    def _stub_raise(self, *args):
-        """Stub outgoing command function (raises an error)."""
-        raise NotImplementedError("Not supported on Clientbot")
-    kill = knock = numeric = _stub_raise
-
-    def update_client(self, target, field, text):
-        """Updates the known ident, host, or realname of a client."""
-        if target not in self.users:
-            log.warning("(%s) Unknown target %s for update_client()", self.name, target)
-            return
-
-        u = self.users[target]
-
-        if field == 'IDENT' and u.ident != text:
-            u.ident = text
-            if not self.is_internal_client(target):
-                # We're updating the host of an external client in our state, so send the appropriate
-                # hook payloads.
-                self.call_hooks([self.sid, 'CHGIDENT',
-                                   {'target': target, 'newident': text}])
-        elif field == 'HOST' and u.host != text:
-            u.host = text
-            if not self.is_internal_client(target):
-                self.call_hooks([self.sid, 'CHGHOST',
-                                   {'target': target, 'newhost': text}])
-        elif field in ('REALNAME', 'GECOS') and u.realname != text:
-            u.realname = text
-            if not self.is_internal_client(target):
-                self.call_hooks([self.sid, 'CHGNAME',
-                                   {'target': target, 'newgecos': text}])
-        else:
-            return  # Nothing changed
-
-    def _get_UID(self, nick, ident=None, host=None):
-        """
-        Fetches the UID for the given nick, creating one if it does not already exist.
-
-        Limited (internal) nick collision checking is done here to prevent Clientbot users from
-        being confused with virtual clients, and vice versa."""
-        self._check_puid_collision(nick)
-
-        idsource = self.nick_to_uid(nick)
-
-        if self.is_internal_client(idsource) and self.pseudoclient and idsource != self.pseudoclient.uid:
-            # We got a message from a client with the same nick as an internal client.
-            # Fire a virtual nick collision to prevent mixing senders.
-            log.debug('(%s) Nick-colliding virtual client %s/%s', self.name, idsource, nick)
-            self.call_hooks([self.sid, 'SAVE', {'target': idsource}])
-
-            # Clear the UID for this nick and spawn a new client for the nick that was just freed.
-            idsource = None
-
-        if idsource is None:
-            # If this sender doesn't already exist, spawn a new client.
-            idsource = self.spawn_client(nick, ident or 'unknown', host or 'unknown',
-                                        server=self.uplink, realname=FALLBACK_REALNAME).uid
-        return idsource
-
-    def parse_message_tags(self, data):
-        """
-        Parses a message with IRC v3.2 message tags, as described at http://ircv3.net/specs/core/message-tags-3.2.html
-        """
-        # Example query:
-        # @aaa=bbb;ccc;example.com/ddd=eee :nick!ident@host.com PRIVMSG me :Hello
-        if data[0].startswith('@'):
-            tagdata = data[0].lstrip('@').split(';')
-            for idx, tag in enumerate(tagdata):
-                tag = tag.replace(r'\s', ' ')
-                tag = tag.replace(r'\\', '\\')
-                tag = tag.replace(r'\r', '\r')
-                tag = tag.replace(r'\n', '\n')
-                tag = tag.replace(r'\:', ';')
-                tagdata[idx] = tag
-
-            results = self.parse_isupport(tagdata, fallback=None)
-            log.debug('(%s) parsed message tags %s', self.name, results)
-            return results
-        return {}
-
     def _set_account_name(self, uid, account):
         """
         Updates the user's account metadata.
@@ -458,7 +484,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
                 except ValueError:
                     ident = host = None  # Set ident and host as null for now.
                     nick = sender  # Treat the sender prefix we received as a nick.
-                idsource = self._get_UID(nick, ident, host)
+                idsource = self._get_UID(nick, ident, host, spawn_new=True)
 
         if idsource in self.users:
             # Handle IRCv3.2 account-tag.
@@ -681,18 +707,26 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         prefix_to_mode = {v:k for k, v in self.prefixmodes.items()}
         prefixes = ''.join(self.prefixmodes.values())
 
-        for name in args[-1].split():
+        # N.B. only split on spaces because of color in hosts nonsense...
+        # str.split() by default treats \x1f as whitespace
+        for name in args[-1].strip().split(' '):
             nick = name.lstrip(prefixes)
 
             # Handle userhost-in-names where available.
+            ident = host = None
             if 'userhost-in-names' in self.ircv3_caps:
-                nick, ident, host = utils.split_hostmask(nick)
-            else:
-                ident = host = None
+                try:
+                    nick, ident, host = utils.split_hostmask(nick)
+                except ValueError:
+                    log.exception('(%s) Failed to split hostmask %r from /names reply on %s; args=%s', self.name, nick, channel, args)
+                    # If error, leave nick unsplit
+
+            if not nick:
+                continue
 
             # Get the PUID for the given nick. If one doesn't exist, spawn
             # a new virtual user.
-            idsource = self._get_UID(nick, ident=ident, host=host)
+            idsource = self._get_UID(nick, ident=ident, host=host, spawn_new=True)
 
             # Queue these virtual users to be joined if they're not already in the channel,
             # or we're waiting for a kick acknowledgment for them.
@@ -720,19 +754,12 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         # Send JOIN hook payloads only for users that we know the ident@host of already.
         # This is mostly used to resync kicked Clientbot users that can't actually be kicked
         # after a delay.
-        if names and hasattr(self.irc.channels[channel], '_clientbot_initial_who_received'):
+        if names and hasattr(self.channels[channel], '_clientbot_initial_who_received'):
             log.debug('(%s) handle_353: sending JOIN hook because /WHO was already received for %s',
-                      self.irc.name, channel)
+                      self.name, channel)
             return {'channel': channel, 'users': names, 'modes': self._channels[channel].modes,
                     'parse_as': "JOIN"}
 
-    def _check_puid_collision(self, nick):
-        """
-        Checks to make sure a nick doesn't clash with a PUID.
-        """
-        if nick in self.users or nick in self.servers:
-            raise ProtocolError("Got bad nick %s from IRC which clashes with a PUID. Is someone trying to spoof users?" % nick)
-
     def _send_who(self, channel):
         """Sends /WHO to a channel, with WHOX args if that is supported."""
         # Note: %% = escaped %
@@ -764,8 +791,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         if command == '352':
             realname = realname.split(' ', 1)[-1]
 
-        self._check_puid_collision(nick)
-        uid = self.nick_to_uid(nick)
+        uid = self._get_UID(nick, spawn_new=False)
 
         if uid is None:
             log.debug("(%s) Ignoring extraneous /WHO info for %s", self.name, nick)
@@ -913,7 +939,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         """
         # <- :GL!~gl@127.0.0.1 KICK #whatever GL| :xd
         channel = args[0]
-        target = self.nick_to_uid(args[1])
+        target = self._get_UID(args[1], spawn_new=False)
 
         try:
             reason = args[2]
@@ -952,7 +978,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         if self.is_channel(target):
             oldobj = self._channels[target].deepcopy()
         else:
-            target = self.nick_to_uid(target)
+            target = self._get_UID(target, spawn_new=False)
             oldobj = None
         modes = args[1:]
         changedmodes = self.parse_modes(target, modes)
@@ -1023,8 +1049,6 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
 
         oldnick = self.users[source].nick
 
-        # Check for any nick collisions with existing virtual clients.
-        self._check_nick_collision(newnick)
         self.users[source].nick = newnick
 
         return {'newnick': newnick, 'oldnick': oldnick}
@@ -1078,7 +1102,7 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
 
         real_target = target.lstrip(''.join(self.prefixmodes.values()))
         if not self.is_channel(real_target):
-            target = self.nick_to_uid(target)
+            target = self._get_UID(target, spawn_new=False)
 
         if target:
             return {'target': target, 'text': args[1]}
@@ -1089,9 +1113,13 @@ class ClientbotWrapperProtocol(IRCCommonProtocol):
         if self.pseudoclient and source == self.pseudoclient.uid:
             # Someone faked a quit from us? We should abort.
             raise ProtocolError("Received QUIT from uplink (%s)" % args[0])
+        elif source not in self.users:
+            log.debug('(%s) Ignoring QUIT on non-existent user %s', self.name, source)
+            return
 
+        userdata = self.users[source]
         self.quit(source, args[0])
-        return {'text': args[0]}
+        return {'text': args[0], 'userdata': userdata}
 
     def handle_404(self, source, command, args):
         """
