@@ -2,13 +2,14 @@
 ircs2s_common.py: Common base protocol class with functions shared by TS6 and P10-based protocols.
 """
 
-import time
 import re
+import time
 from collections import defaultdict
 
+from pylinkirc import conf, utils
 from pylinkirc.classes import IRCNetwork, ProtocolError
 from pylinkirc.log import log
-from pylinkirc import utils, conf
+
 
 class IncrementalUIDGenerator():
     """
@@ -60,6 +61,7 @@ class IRCCommonProtocol(IRCNetwork):
 
         self._caps = {}
         self._use_builtin_005_handling = False  # Disabled by default for greater security
+        self.protocol_caps |= {'has-irc-modes', 'can-manage-bot-channels'}
 
     def post_connect(self):
         self._caps.clear()
@@ -77,30 +79,6 @@ class IRCCommonProtocol(IRCNetwork):
                  "Invalid port %r for network %s"
                  % (port, self.name))
 
-    # TODO: these wrappers really need to be standardized
-    def _get_SID(self, sname):
-        """Returns the SID of a server with the given name, if present."""
-        name = sname.lower()
-
-        if name in self.servers:
-            return name
-
-        for k, v in self.servers.items():
-            if v.name.lower() == name:
-                return k
-        else:
-            return sname  # Fall back to given text instead of None
-
-    def _get_UID(self, target):
-        """Converts a nick argument to its matching UID. This differs from irc.nick_to_uid()
-        in that it returns the original text instead of None, if no matching nick is found."""
-
-        if target in self.users:
-            return target
-
-        target = self.nick_to_uid(target) or target
-        return target
-
     @staticmethod
     def parse_args(args):
         """
@@ -118,7 +96,8 @@ class IRCCommonProtocol(IRCNetwork):
                 joined_arg = ' '.join(args[idx:])[1:]  # Cut off the leading : as well
                 real_args.append(joined_arg)
                 break
-            real_args.append(arg)
+            elif arg.strip():  # Skip empty args that aren't part of the multi-word arg
+                real_args.append(arg)
 
         return real_args
 
@@ -130,63 +109,6 @@ class IRCCommonProtocol(IRCNetwork):
         args[0] = args[0].split(':', 1)[1]
         return args
 
-    def _squit(self, numeric, command, args):
-        """Handles incoming SQUITs."""
-
-        split_server = self._get_SID(args[0])
-
-        # Normally we'd only need to check for our SID as the SQUIT target, but Nefarious
-        # actually uses the uplink server as the SQUIT target.
-        # <- ABAAE SQ nefarious.midnight.vpn 0 :test
-        if split_server in (self.sid, self.uplink):
-            raise ProtocolError('SQUIT received: (reason: %s)' % args[-1])
-
-        affected_users = []
-        affected_nicks = defaultdict(list)
-        log.debug('(%s) Splitting server %s (reason: %s)', self.name, split_server, args[-1])
-
-        if split_server not in self.servers:
-            log.warning("(%s) Tried to split a server (%s) that didn't exist!", self.name, split_server)
-            return
-
-        # Prevent RuntimeError: dictionary changed size during iteration
-        old_servers = self.servers.copy()
-        old_channels = self._channels.copy()
-
-        # Cycle through our list of servers. If any server's uplink is the one that is being SQUIT,
-        # remove them and all their users too.
-        for sid, data in old_servers.items():
-            if data.uplink == split_server:
-                log.debug('Server %s also hosts server %s, removing those users too...', split_server, sid)
-                # Recursively run SQUIT on any other hubs this server may have been connected to.
-                args = self._squit(sid, 'SQUIT', [sid, "0",
-                                   "PyLink: Automatically splitting leaf servers of %s" % sid])
-                affected_users += args['users']
-
-        for user in self.servers[split_server].users.copy():
-            affected_users.append(user)
-            nick = self.users[user].nick
-
-            # Nicks affected is channel specific for SQUIT:. This makes Clientbot's SQUIT relaying
-            # much easier to implement.
-            for name, cdata in old_channels.items():
-                if user in cdata.users:
-                    affected_nicks[name].append(nick)
-
-            log.debug('Removing client %s (%s)', user, nick)
-            self._remove_client(user)
-
-        serverdata = self.servers[split_server]
-        sname = serverdata.name
-        uplink = serverdata.uplink
-
-        del self.servers[split_server]
-        log.debug('(%s) Netsplit affected users: %s', self.name, affected_users)
-
-        return {'target': split_server, 'users': affected_users, 'name': sname,
-                'uplink': uplink, 'nicks': affected_nicks, 'serverdata': serverdata,
-                'channeldata': old_channels}
-
     @staticmethod
     def parse_isupport(args, fallback=''):
         """
@@ -217,6 +139,33 @@ class IRCCommonProtocol(IRCNetwork):
         prefixsearch = re.search(r'\(([A-Za-z]+)\)(.*)', args)
         return dict(zip(prefixsearch.group(1), prefixsearch.group(2)))
 
+    @classmethod
+    def parse_message_tags(cls, data):
+        """
+        Parses IRCv3.2 message tags from a message, as described at http://ircv3.net/specs/core/message-tags-3.2.html
+
+        data is a list of command arguments, split by spaces.
+        """
+        # Example query:
+        # @aaa=bbb;ccc;example.com/ddd=eee :nick!ident@host.com PRIVMSG me :Hello
+        if data[0].startswith('@'):
+            tagdata = data[0].lstrip('@').split(';')
+            for idx, tag in enumerate(tagdata):
+                tag = tag.replace('\\s', ' ')
+                tag = tag.replace('\\r', '\r')
+                tag = tag.replace('\\n', '\n')
+                tag = tag.replace('\\:', ';')
+
+                # We want to drop lone \'s but keep \\ as \ ...
+                tag = tag.replace('\\\\', '\x00')
+                tag = tag.replace('\\', '')
+                tag = tag.replace('\x00', '\\')
+                tagdata[idx] = tag
+
+            results = cls.parse_isupport(tagdata, fallback='')
+            return results
+        return {}
+
     def handle_away(self, source, command, args):
         """Handles incoming AWAY messages."""
         # TS6:
@@ -333,8 +282,8 @@ class IRCS2SProtocol(IRCCommonProtocol):
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
-        self.protocol_caps = {'can-spawn-clients', 'has-ts', 'can-host-relay',
-                              'can-track-servers'}
+        self.protocol_caps |= {'can-spawn-clients', 'has-ts', 'can-host-relay',
+                               'can-track-servers'}
 
         # Alias
         self.handle_squit = self._squit
@@ -351,6 +300,12 @@ class IRCS2SProtocol(IRCCommonProtocol):
         the SID of the uplink server.
         """
         data = data.split(" ")
+
+        tags = self.parse_message_tags(data)
+        if tags:
+            # If we have message tags, split off the first argument.
+            data = data[1:]
+
         args = self.parse_args(data)
 
         sender = args[0]
@@ -399,10 +354,12 @@ class IRCS2SProtocol(IRCCommonProtocol):
         else:
             parsed_args = func(sender, command, args)
             if parsed_args is not None:
+                if tags:
+                    parsed_args['tags'] = tags  # Add message tags to this hook payload.
                 return [sender, command, parsed_args]
 
     def invite(self, source, target, channel):
-        """Sends an INVITE from a PyLink client.."""
+        """Sends an INVITE from a PyLink client."""
         if not self.is_internal_client(source):
             raise LookupError('No such PyLink client exists.')
 
@@ -451,7 +408,7 @@ class IRCS2SProtocol(IRCCommonProtocol):
 
         This is mostly used by PyLink internals to check whether the remote link is up."""
         if self.sid and self.connected.is_set():
-            self._send_with_prefix(self.sid, 'PING %s' % self._expandPUID(self.sid))
+            self._send_with_prefix(self.sid, 'PING %s' % self._expandPUID(self.uplink))
 
     def quit(self, numeric, reason):
         """Quits a PyLink client."""
@@ -538,14 +495,15 @@ class IRCS2SProtocol(IRCCommonProtocol):
     def handle_kill(self, source, command, args):
         """Handles incoming KILLs."""
         killed = self._get_UID(args[0])
-        # Depending on whether the IRCd sends explicit QUIT messages for
-        # killed clients, the user may or may not have automatically been
-        # removed from our user list.
-        # If not, we have to assume that KILL = QUIT and remove them
-        # ourselves.
-        data = self.users.get(killed)
-        if data:
-            self._remove_client(killed)
+        # Some IRCds send explicit QUIT messages for their killed clients in addition to KILL,
+        # meaning that our target client may have been removed already. If this is the case,
+        # don't bother forwarding this message on.
+        # Generally, we only need to distinguish between KILL and QUIT if the target is
+        # one of our clients, in which case the above statement isn't really applicable.
+        if killed in self.users:
+            userdata = self._remove_client(killed)
+        else:
+            return
 
         # TS6-style kills look something like this:
         # <- :GL KILL 38QAAAAAA :hidden-1C620195!GL (test)
@@ -564,8 +522,8 @@ class IRCS2SProtocol(IRCCommonProtocol):
             # Get the reason, which is enclosed in brackets.
             killmsg = ' '.join(args[1].split(" ")[1:])[1:-1]
             if not killmsg:
-                log.warning('(%s) Failed to extract kill reason: %r', irc.name, args)
-                killmsg = '<No reason given>'
+                log.warning('(%s) Failed to extract kill reason: %r', self.name, args)
+                killmsg = args[1]
         else:
             # We already have a preformatted kill, so just pass it on as is.
             # XXX: this does create a convoluted kill string if we want to forward kills
@@ -576,9 +534,9 @@ class IRCS2SProtocol(IRCCommonProtocol):
             # <- :GL KILL PyLink-devel :KILLed by GL: ?
             killmsg = args[1]
 
-        return {'target': killed, 'text': killmsg, 'userdata': data}
+        return {'target': killed, 'text': killmsg, 'userdata': userdata}
 
-    def _check_cloak_change(self, uid):
+    def _check_cloak_change(self, uid):  # Stub by default
         return
 
     def _check_umode_away_change(self, uid):
@@ -726,8 +684,9 @@ class IRCS2SProtocol(IRCCommonProtocol):
         # <- :1SRAAGB4T QUIT :Quit: quit message goes here
         # P10:
         # <- ABAAB Q :Killed (GL_ (bangbang))
-        self._remove_client(numeric)
-        return {'text': args[0]}
+        userdata = self._remove_client(numeric)
+        if userdata:
+            return {'text': args[0], 'userdata': userdata}
 
     def handle_stats(self, numeric, command, args):
         """Handles the IRC STATS command."""
