@@ -7,14 +7,15 @@ ngircd.py: PyLink protocol module for ngIRCd.
 # and https://tools.ietf.org/html/rfc2813
 ##
 
-import time
 import re
+import time
 
-from pylinkirc import utils, conf, __version__
+from pylinkirc import __version__, conf, utils
 from pylinkirc.classes import *
 from pylinkirc.log import log
 from pylinkirc.protocols.ircs2s_common import *
 
+
 class NgIRCdProtocol(IRCS2SProtocol):
     def __init__(self, irc):
         super().__init__(irc)
@@ -58,15 +59,31 @@ class NgIRCdProtocol(IRCS2SProtocol):
         self._caps.clear()
 
         self.cmodes.update({
-            'banexception': 'e', 'invex': 'I', 'regmoderated': 'M', 'nonick': 'N',
-            'operonly': 'O', 'permanent': 'P', 'nokick': 'Q', 'registered': 'r',
-            'regonly': 'R', 'noinvite': 'V', 'sslonly': 'z'
+            'banexception': 'e',
+            'invex': 'I',
+            'noinvite': 'V',
+            'nokick': 'Q',
+            'nonick': 'N',
+            'operonly': 'O',
+            'permanent': 'P',
+            'registered': 'r',
+            'regmoderated': 'M',
+            'regonly': 'R',
+            'sslonly': 'z'
         })
 
         self.umodes.update({
-            'away': 'a', 'deaf': 'b', 'bot': 'B', 'sno_clientconnections': 'c',
-            'deaf_commonchan': 'C', 'floodexempt': 'f', 'hidechans': 'I',
-            'servprotect': 'q', 'restricted': 'r', 'registered': 'R', 'cloak': 'x'
+            'away': 'a',
+            'bot': 'B',
+            'cloak': 'x',
+            'deaf_commonchan': 'C',
+            'floodexempt': 'F',
+            'hidechans': 'I',
+            'privdeaf': 'b',
+            'registered': 'R',
+            'restricted': 'r',
+            'servprotect': 'q',
+            'sno_clientconnections': 'c'
         })
 
     def spawn_client(self, nick, ident='null', host='null', realhost=None, modes=set(),
@@ -510,7 +527,8 @@ class NgIRCdProtocol(IRCS2SProtocol):
         if recvpass != self.serverdata['recvpass']:
             raise ProtocolError("RECVPASS from uplink does not match configuration!")
 
-        assert 'IRC+' in args[1], "Linking to non-ngIRCd server using this protocol module is not supported"
+        if 'IRC+' not in args[1]:
+            raise ProtocolError("Linking to non-ngIRCd server using this protocol module is not supported")
 
     def handle_ping(self, source, command, args):
         """
