@@ -4,14 +4,15 @@ p10.py: P10 protocol module for PyLink, supporting Nefarious IRCu and others.
 
 import base64
 import struct
-from ipaddress import ip_address
 import time
+from ipaddress import ip_address
 
-from pylinkirc import utils, structures, conf
+from pylinkirc import conf, structures, utils
 from pylinkirc.classes import *
 from pylinkirc.log import log
 from pylinkirc.protocols.ircs2s_common import *
 
+
 class P10UIDGenerator(IncrementalUIDGenerator):
      """Implements an incremental P10 UID Generator."""
 
@@ -1274,11 +1275,16 @@ class P10Protocol(IRCS2SProtocol):
 
             # Any other subcommands listed at https://github.com/evilnet/nefarious2/blob/master/doc/p10.txt#L354
             # shouldn't apply to us.
-
             if args[1] in ('R', 'M'):
                 accountname = args[2]
             elif args[1] == 'U':
                 accountname = ''  # logout
+            elif len(args[1]) > 1:
+                log.warning('(%s) Got subcommand %r for %s in ACCOUNT message, is use_extended_accounts set correctly?',
+                            self.name, args[1], target)
+                return
+            else:
+                return
 
         else:
             # ircu or nefarious with F:EXTENDED_ACCOUNTS = FALSE
