@@ -2,14 +2,15 @@
 ts6.py: PyLink protocol module for TS6-based IRCds (charybdis, elemental-ircd).
 """
 
-import time
 import re
+import time
 
-from pylinkirc import utils, conf
+from pylinkirc import conf, utils
 from pylinkirc.classes import *
 from pylinkirc.log import log
 from pylinkirc.protocols.ts6_common import *
 
+
 class TS6Protocol(TS6BaseProtocol):
 
     SUPPORTED_IRCDS = ('charybdis', 'elemental', 'chatircd', 'ratbox')
@@ -90,6 +91,19 @@ class TS6Protocol(TS6BaseProtocol):
 
         return u
 
+    def spawn_server(self, name, sid=None, uplink=None, desc=None):
+        """
+        Spawns a server off a PyLink server. desc (server description)
+        defaults to the one in the config. uplink defaults to the main PyLink
+        server, and sid (the server ID) is automatically generated if not
+        given.
+        """
+        desc = desc or self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']
+        if self.serverdata.get('hidden', False):
+            desc = '(H) ' + desc
+
+        return super().spawn_server(name, sid=sid, uplink=uplink, desc=desc)
+
     def join(self, client, channel):
         """Joins a PyLink client to a channel."""
         # JOIN:
@@ -361,7 +375,7 @@ class TS6Protocol(TS6BaseProtocol):
             self.cmodes.update(chatircd_cmodes)
             self.cmodes['*D'] += ''.join(chatircd_cmodes.values())
 
-            chatircd_umodes = {'netadmin': 'n', 'bot': 'B', 'callerid_sslonly': 't'}
+            chatircd_umodes = {'netadmin': 'n', 'bot': 'B', 'sslonlymsg': 't'}
             self.umodes.update(chatircd_umodes)
             self.umodes['*D'] += ''.join(chatircd_umodes.values())
 
@@ -397,8 +411,10 @@ class TS6Protocol(TS6BaseProtocol):
         # KLN: supports remote KLINEs
         f('CAPAB :QS ENCAP EX CHW IE KNOCK SAVE SERVICES TB EUID RSFNC EOPMOD SAVETS_100 KLN')
 
-        f('SERVER %s 0 :%s' % (self.serverdata["hostname"],
-                               self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']))
+        sdesc = self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']
+        if self.serverdata.get('hidden', False):
+            sdesc = '(H) ' + sdesc
+        f('SERVER %s 0 :%s' % (self.serverdata["hostname"], sdesc))
 
         # Finally, end all the initialization with a PING - that's Charybdis'
         # way of saying end-of-burst :)
