@@ -1,13 +1,13 @@
 """
-unreal.py: UnrealIRCd 4.x protocol module for PyLink.
+unreal.py: UnrealIRCd 4.x-5.x protocol module for PyLink.
 """
 
-import time
 import codecs
-import socket
 import re
+import socket
+import time
 
-from pylinkirc import utils, conf
+from pylinkirc import conf, utils
 from pylinkirc.classes import *
 from pylinkirc.log import log
 from pylinkirc.protocols.ts6_common import *
@@ -20,6 +20,57 @@ class UnrealProtocol(TS6BaseProtocol):
     # https://github.com/unrealircd/unrealircd/blob/4cad9cb/src/modules/m_server.c#L1260 may
     # also help. (but why BUFSIZE-*80*?) -GL
     S2S_BUFSIZE = 427
+    _KNOWN_CMODES = {'ban': 'b',
+              'banexception': 'e',
+              'blockcolor': 'c',
+              'censor': 'G',
+              'delayjoin': 'D',
+              'flood_unreal': 'f',
+              'invex': 'I',
+              'inviteonly': 'i',
+              'issecure': 'Z',
+              'key': 'k',
+              'limit': 'l',
+              'moderated': 'm',
+              'noctcp': 'C',
+              'noextmsg': 'n',
+              'noinvite': 'V',
+              'nokick': 'Q',
+              'noknock': 'K',
+              'nonick': 'N',
+              'nonotice': 'T',
+              'op': 'o',
+              'operonly': 'O',
+              'permanent': 'P',
+              'private': 'p',
+              'registered': 'r',
+              'regmoderated': 'M',
+              'regonly': 'R',
+              'secret': 's',
+              'sslonly': 'z',
+              'stripcolor': 'S',
+              'topiclock': 't',
+              'voice': 'v'}
+    _KNOWN_UMODES = {'bot': 'B',
+              'censor': 'G',
+              'cloak': 'x',
+              'deaf': 'd',
+              'filter': 'G',
+              'hidechans': 'p',
+              'hideidle': 'I',
+              'hideoper': 'H',
+              'invisible': 'i',
+              'noctcp': 'T',
+              'protected': 'q',
+              'regdeaf': 'R',
+              'registered': 'r',
+              'sslonlymsg': 'Z',
+              'servprotect': 'S',
+              'showwhois': 'W',
+              'snomask': 's',
+              'ssl': 'z',
+              'vhost': 't',
+              'wallops': 'w'}
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -28,7 +79,7 @@ class UnrealProtocol(TS6BaseProtocol):
         self.casemapping = 'ascii'
 
         # Unreal protocol version
-        self.proto_ver = 4200
+        self.proto_ver = 4203
         self.min_proto_ver = 4000
 
         self.hook_map = {'UMODE2': 'MODE', 'SVSKILL': 'KILL', 'SVSMODE': 'MODE',
@@ -341,15 +392,6 @@ class UnrealProtocol(TS6BaseProtocol):
         # Track usages of legacy (Unreal 3.2) nicks.
         self.legacy_uidgen = PUIDGenerator('U32user')
 
-        self.umodes.update({'deaf': 'd', 'invisible': 'i', 'hidechans': 'p',
-                                'protected': 'q', 'registered': 'r',
-                                'snomask': 's', 'vhost': 't', 'wallops': 'w',
-                                'bot': 'B', 'cloak': 'x', 'ssl': 'z',
-                                'filter': 'G', 'hideoper': 'H', 'hideidle': 'I',
-                                'regdeaf': 'R', 'servprotect': 'S',
-                                'noctcp': 'T', 'showwhois': 'W',
-                                '*A': '', '*B': '', '*C': '', '*D': 'dipqrstwBxzGHIRSTW'})
-
         f = self.send
         host = self.serverdata["hostname"]
 
@@ -375,14 +417,28 @@ class UnrealProtocol(TS6BaseProtocol):
         f('PROTOCTL SJOIN SJ3 NOQUIT NICKv2 VL UMODE2 PROTOCTL NICKIP EAUTH=%s SID=%s VHP ESVID' % (self.serverdata["hostname"], self.sid))
         sdesc = self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']
         f('SERVER %s 1 U%s-h6e-%s :%s' % (host, self.proto_ver, self.sid, sdesc))
-        f('NETINFO 1 %s %s * 0 0 0 :%s' % (self.start_ts, self.proto_ver, self.serverdata.get("netname", self.name)))
+
         self._send_with_prefix(self.sid, 'EOS')
 
         # Extban definitions
-        self.extbans_acting = {'quiet': '~q:', 'ban_nonick': '~n:', 'ban_nojoins': '~j:',
-                               'filter': '~T:block:', 'filter_censor': '~T:censor:'}
-        self.extbans_matching = {'ban_account': '~a:', 'ban_inchannel': '~c:', 'ban_opertype': '~O:',
-                                 'ban_realname': '~r:', 'ban_account_legacy': '~R:', 'ban_certfp': '~S:'}
+        self.extbans_acting = {'quiet': '~q:',
+                               'ban_nonick': '~n:',
+                               'ban_nojoins': '~j:',
+                               'filter': '~T:block:',
+                               'filter_censor': '~T:censor:',
+                               'msgbypass_external': '~m:external:',
+                               'msgbypass_censor': '~m:censor:',
+                               'msgbypass_moderated': '~m:moderated:',
+                               # These two sort of map to InspIRCd +e S: and +e T:
+                               'ban_stripcolor': '~m:color:',
+                               'ban_nonotice': '~m:notice:',
+                               'timedban_unreal': '~t:'}
+        self.extbans_matching = {'ban_account': '~a:',
+                                 'ban_inchannel': '~c:',
+                                 'ban_opertype': '~O:',
+                                 'ban_realname': '~r:',
+                                 'ban_account_legacy': '~R:',
+                                 'ban_certfp': '~S:'}
 
     def handle_eos(self, numeric, command, args):
         """EOS is used to denote end of burst."""
@@ -496,6 +552,10 @@ class UnrealProtocol(TS6BaseProtocol):
                                     "(Unreal 4.x), got %s)" % (self.min_proto_ver, protover))
             self.servers[numeric] = Server(self, None, sname, desc=sdesc)
 
+            # Prior to 4203, Unreal did not send PROTOCTL USERMODES (see handle_protoctl() )
+            if protover < 4203:
+                self.umodes.update(self._KNOWN_UMODES)
+                self.umodes['*D'] = ''.join(self._KNOWN_UMODES.values())
         else:
             # Legacy (non-SID) servers can still be introduced using the SERVER command.
             # <- :services.int SERVER a.bc 2 :(H) [GL] a
@@ -503,21 +563,19 @@ class UnrealProtocol(TS6BaseProtocol):
 
     def handle_protoctl(self, numeric, command, args):
         """Handles protocol negotiation."""
-
-        cmodes = {'noknock': 'K', 'limit': 'l', 'registered': 'r', 'flood_unreal': 'f',
-                  'censor': 'G', 'noextmsg': 'n', 'invex': 'I', 'permanent': 'P',
-                  'sslonly': 'z', 'operonly': 'O', 'moderated': 'm', 'blockcolor': 'c',
-                  'regmoderated': 'M', 'noctcp': 'C', 'secret': 's', 'ban': 'b',
-                  'nokick': 'Q', 'private': 'p', 'stripcolor': 'S', 'key': 'k',
-                  'op': 'o', 'voice': 'v', 'regonly': 'R', 'noinvite': 'V',
-                  'banexception': 'e', 'nonick': 'N', 'issecure': 'Z', 'topiclock': 't',
-                  'nonotice': 'T', 'delayjoin': 'D', 'inviteonly': 'i'}
-
         # Make a list of all our capability names.
         self.caps += [arg.split('=')[0] for arg in args]
 
+        # Unreal 4.0.x:
         # <- PROTOCTL NOQUIT NICKv2 SJOIN SJOIN2 UMODE2 VL SJ3 TKLEXT TKLEXT2 NICKIP ESVID
         # <- PROTOCTL CHANMODES=beI,k,l,psmntirzMQNRTOVKDdGPZSCc NICKCHARS= SID=001 MLOCK TS=1441314501 EXTSWHOIS
+        # Unreal 4.2.x:
+        # <- PROTOCTL NOQUIT NICKv2 SJOIN SJOIN2 UMODE2 VL SJ3 TKLEXT TKLEXT2 NICKIP ESVID SJSBY
+        # <- PROTOCTL CHANMODES=beI,kLf,l,psmntirzMQNRTOVKDdGPZSCc USERMODES=iowrsxzdHtIDZRqpWGTSB BOOTED=1574014839 PREFIX=(qaohv)~&@%+ NICKCHARS= SID=001 MLOCK TS=1574020869 EXTSWHOIS
+        # Unreal 5.0.0-rc1:
+        # <- PROTOCTL NOQUIT NICKv2 SJOIN SJOIN2 UMODE2 VL SJ3 TKLEXT TKLEXT2 NICKIP ESVID SJSBY MTAGS
+        # <- PROTOCTL CHANMODES=beI,kLf,lH,psmntirzMQNRTOVKDdGPZSCc USERMODES=iowrsxzdHtIDZRqpWGTSB BOOTED=1574020755 PREFIX=(qaohv)~&@%+ SID=001 MLOCK TS=1574020823 EXTSWHOIS
+        # <- PROTOCTL NICKCHARS= CHANNELCHARS=utf8
         for cap in args:
             if cap.startswith('SID'):
                 self.uplink = cap.split('=', 1)[1]
@@ -525,14 +583,18 @@ class UnrealProtocol(TS6BaseProtocol):
                 # Parse all the supported channel modes.
                 supported_cmodes = cap.split('=', 1)[1]
                 self.cmodes['*A'], self.cmodes['*B'], self.cmodes['*C'], self.cmodes['*D'] = supported_cmodes.split(',')
-                for namedmode, modechar in cmodes.items():
+                for namedmode, modechar in self._KNOWN_CMODES.items():
                     if modechar in supported_cmodes:
                         self.cmodes[namedmode] = modechar
-                self.cmodes['*B'] += 'f'  # Add +f to the list too, dunno why it isn't there.
+            elif cap.startswith('USERMODES'):  # Only for protover >= 4203
+                self.umodes['*D'] = supported_umodes = cap.split('=', 1)[1]
+                for namedmode, modechar in self._KNOWN_UMODES.items():
+                    if modechar in supported_umodes:
+                        self.umodes[namedmode] = modechar
 
         # Add in the supported prefix modes.
         self.cmodes.update({'halfop': 'h', 'admin': 'a', 'owner': 'q',
-                                'op': 'o', 'voice': 'v'})
+                            'op': 'o', 'voice': 'v'})
 
     def handle_join(self, numeric, command, args):
         """Handles the UnrealIRCd JOIN command."""
@@ -607,6 +669,11 @@ class UnrealProtocol(TS6BaseProtocol):
                     continue
 
                 user = self._get_UID(user)  # Normalize nicks to UIDs for Unreal 3.2 links
+                if user not in self.users:
+                    # Work around a potential race when sending kills on join
+                    log.debug("(%s) Ignoring user %s in SJOIN to %s, they don't exist anymore", self.name, user, channel)
+                    continue
+
                 # Unreal uses slightly different prefixes in SJOIN. +q is * instead of ~,
                 # and +a is ~ instead of &.
                 modeprefix = (r.group(1) or '').replace("~", "&").replace("*", "~")
