@@ -5,19 +5,20 @@ This module contains various utility functions related to IRC and/or the PyLink
 framework.
 """
 
-import string
-import re
-import importlib
-import os
-import collections
 import argparse
+import collections
+import functools
+import importlib
 import ipaddress
-
-from .log import log
-from . import world, conf, structures
+import os
+import re
+import string
 
 # Load the protocol and plugin packages.
-from pylinkirc import protocols, plugins
+from pylinkirc import plugins, protocols
+
+from . import conf, structures, world
+from .log import log
 
 PLUGIN_PREFIX = plugins.__name__ + '.'
 PROTOCOL_PREFIX = protocols.__name__ + '.'
@@ -92,6 +93,8 @@ def split_hostmask(mask):
     """
     nick, identhost = mask.split('!', 1)
     ident, host = identhost.split('@', 1)
+    if not all({nick, ident, host}):
+        raise ValueError("Invalid user@host %r" % mask)
     return [nick, ident, host]
 splitHostmask = split_hostmask
 
@@ -150,7 +153,7 @@ class ServiceBot():
         else:
             raise NotImplementedError("Network specific plugins not supported yet.")
 
-    def join(self, irc, channels, ignore_empty=True):
+    def join(self, irc, channels, ignore_empty=None):
         """
         Joins the given service bot to the given channel(s). "channels" can be
         an iterable of channel names or the name of a single channel (type 'str').
@@ -160,8 +163,8 @@ class ServiceBot():
         marked persistent). This option is automatically *disabled* on networks
         where we cannot monitor channels that we're not in (e.g. Clientbot).
 
-        Before PyLink 2.0-alpha3, this function implicitly marks channels i
-        receives to be persistent - this is no longer the case!
+        Before PyLink 2.0-alpha3, this function implicitly marked channels it
+        receives to be persistent. This behaviour is no longer the case.
         """
         uid = self.uids.get(irc.name)
         if uid is None:
@@ -173,6 +176,10 @@ class ServiceBot():
         if irc.has_cap('visible-state-only'):
             # Disable dynamic channel joining on networks where we can't monitor channels for joins.
             ignore_empty = False
+        elif ignore_empty is None:
+            ignore_empty = not (irc.serverdata.get('join_empty_channels',
+                                                   conf.conf['pylink'].get('join_empty_channels',
+                                                                           False)))
 
         # Specify modes to join the services bot with.
         joinmodes = irc.get_service_option(self.name, 'joinmodes', default='')
@@ -392,7 +399,7 @@ class ServiceBot():
         chanlist = namespace.setdefault(irc.name, structures.IRCCaseInsensitiveSet(irc))
         chanlist.add(channel)
 
-        if try_join:
+        if try_join and irc.has_cap('can-manage-bot-channels'):
             self.join(irc, [channel])
 
     def remove_persistent_channel(self, irc, namespace, channel, try_part=True, part_reason=''):
@@ -401,7 +408,7 @@ class ServiceBot():
         """
         chanlist = self.dynamic_channels[namespace][irc.name].remove(channel)
 
-        if try_part and irc.connected.is_set():
+        if try_part and irc.connected.is_set() and irc.has_cap('can-manage-bot-channels'):
             self.part(irc, [channel], reason=part_reason)
 
     def get_persistent_channels(self, irc, namespace=None):
@@ -817,3 +824,48 @@ def parse_duration(text):
         raise ValueError("Failed to parse duration string %r" % text)
 
     return result
+
+@functools.lru_cache(maxsize=1024)
+def _glob2re(glob):
+    """Converts an IRC-style glob to a regular expression."""
+    patt = ['^']
+
+    for char in glob:
+        if char == '*' and patt[-1] != '*':  # Collapse ** into *
+            patt.append('.*')
+        elif char == '?':
+            patt.append('.')
+        else:
+            patt.append(re.escape(char))
+
+    patt.append('$')
+    return ''.join(patt)
+
+def match_text(glob, text, filterfunc=str.lower):
+    """
+    Returns whether glob matches text. If filterfunc is specified, run filterfunc on glob and text
+    before preforming matches.
+    """
+    if filterfunc:
+        glob = filterfunc(glob)
+        text = filterfunc(text)
+
+    return re.match(_glob2re(glob), text)
+
+def merge_iterables(A, B):
+    """
+    Merges the values in two iterables. A and B must be of the same type, and one of the following:
+
+    - list: items are combined as A + B
+    - set:  items are combined as A | B
+    - dict: items are combined as {**A, **B}
+    """
+    if type(A) != type(B):
+        raise ValueError("inputs must be the same type")
+
+    if isinstance(A, list):
+        return A + B
+    elif isinstance(A, set):
+        return A | B
+    elif isinstance(A, dict):
+        return {**A, **B}
